from typing import Union, Optional, List
import io
import time
import traceback

import pandas as pd

from .. import api
from ._dtos import (
    get_default_model_execution,
    get_default_execution_report,
    ExecutionReport,
    CustomModelOutput,
)
from ._queries import update_instance_resource

__all__ = [
    "FileOutput",
    "InstanceResourceOutput",
    "TimeValuesOutput",
    "VectorTimeValuesOutput",
    "Delay",
    "BatchValuesOutput",
    "VectorBatchValuesOutput",
    "BatchFeaturesOutput",
    "BatchesOutput",
    "CustomTextOutput",
    "CustomJsonOutput",
    "BatchComputationJob",
    "OIModelOutputs",
]


# Output classes
class FileOutput:
    """
    This class contains the content to be sent in a file upload to Oi Analytics.

    Attributes
    ----------
    output_type : {'file'}
        Denotes the specific type of the current output object.
    file_name : str
        Name of the output file.
    content : io.StringIO | io.BytesIO
        Content of the output file.

    Methods
    -------
    from_pandas(self, data, file_name, file_type):
        Initialize an instance of FileOutput with content from pandas.Series or pandas.DataFrame.

    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
        Upload file to Oi Analytics.
    """

    def __init__(
        self,
        file_name: str,
        content: Union[io.StringIO, io.BytesIO],
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """
        Initialize a new instance of FileOutput.

        Parameters
        ----------
        file_name : str
            Name of the output file.
        content : io.StringIO | io.BytesIO
            Content of the output file.
        """
        self.output_type = "file"
        self.file_name = file_name
        self.content = content
        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    @classmethod
    def from_pandas(
        cls,
        data: Union[pd.Series, pd.DataFrame],
        file_name: str,
        file_type: str = "csv",
        writing_kwargs: Optional[dict] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """
        Initialize an instance of FileOutput with data from pandas.Series or pandas.DataFrame.

        Parameters
        ----------
        data : pandas.Series | pandas.DataFrame
            Output data.
        file_name : str
            Name of the output file.
        file_type : str, default 'csv'
            Extension of the output file.
        writing_kwargs : dict, optional

        Returns
        -------
        object
            FileOutput.

        """
        # Init
        if writing_kwargs is None:
            writing_kwargs = {}

        bio = io.BytesIO()

        # Write data
        if file_type == "excel":
            data.to_excel(bio, **writing_kwargs)
        elif file_type == "csv":
            data.to_csv(bio, **writing_kwargs)
        else:
            raise NotImplementedError(f"Unsupported file_type: {file_type}")
        bio.seek(0)

        # Create object
        return cls(
            file_name=file_name,
            content=bio,
            raise_exceptions_on_send=raise_exceptions_on_send,
            log_exceptions_if_not_raised=log_exceptions_if_not_raised,
        )

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        Upload file to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used
        execution_report : ExecutionReport, optional
            Execution report for external use
        raise_exceptions : bool, default False
            Whether to raise exception
        log_exceptions_if_not_raised : bool, default True
            Whether to print exception

        Returns
        -------
        None

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # get execution report
        if execution_report is None:
            execution_report = get_default_execution_report()
        # update execution report
        try:
            response = api.endpoints.files.upload_file(
                file_content=self.content,
                file_name=self.file_name,
                api_credentials=api_credentials,
            )
            execution_report.update(files_uploaded=1)
            return response
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class InstanceResourceOutput:
    """
    This class contains the content to be sent as an instance resource to Oi Analytics.

    Attributes
    ----------
    output_type : 'instance_resource'
        Denotes the specific type of the current output object.
    file_content : file_content
        Content of the instance resource output file.
    file_name : str
        Name of the instance resource output file.
    resource_file_id : resource_file_id
        Unique identifier of instance resource output file.
    model_instance_id : model_instance_id
        Unique identifier of model instance.

    Methods
    -------
    from_pandas(self, data, file_name, file_type):
        Create InstanceResourceOutput object from a pandas.Series or a pandas.DataFrame.

    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
        Upload instance resource to Oi Analytics.
    """

    def __init__(
        self,
        file_content: Union[io.StringIO, io.BytesIO],
        file_name: str,
        resource_file_id: str,
        model_instance_id: Optional[str] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """
        Initialize a new instance of InstanceResourceOutput.

        Parameters
        ----------
        file_content : io.StringIO | io.BytesIO
            Content of the instance resource output file.
        file_name : str
            Name of the instance resource output file.
        resource_file_id : str
            Unique identifier of instance resource output file.
        model_instance_id : str, optional
            Unique identifier of model instance.
        """
        self.output_type = "instance_resource"
        self.file_content = file_content
        self.file_name = file_name
        self.resource_file_id = resource_file_id
        self.model_instance_id = model_instance_id
        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    @classmethod
    def from_pandas(
        cls,
        data: Union[pd.Series, pd.DataFrame],
        file_name: str,
        resource_file_id: str,
        model_instance_id: Optional[str] = None,
        file_type: str = "csv",
        writing_kwargs: Optional[dict] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """Create InstanceResourceOutput object from a pandas.Series or a pandas.DataFrame.

        Parameters
        ----------
        data : pandas.Series, pandas.DataFrame
            Data of instance resource output.
        file_name : str
            Name of instance resource output file.
        resource_file_id : str
            ID of insta,ce resource output file.
        model_instance_id : str, optional
            ID of model instance.
        file_type : str, default 'csv'
            Extension of the output file.
        writing_kwargs : dict, optional

        Returns
        -------
        object
            InstanceResourceOutput

        """
        # Init
        if writing_kwargs is None:
            writing_kwargs = {}

        bio = io.BytesIO()

        # Write data
        if file_type == "excel":
            data.to_excel(bio, **writing_kwargs)
        elif file_type == "csv":
            data.to_csv(bio, **writing_kwargs)
        else:
            raise NotImplementedError(f"Unsupported file_type: {file_type}")
        bio.seek(0)

        # Create object
        return cls(
            file_content=bio,
            file_name=file_name,
            resource_file_id=resource_file_id,
            model_instance_id=model_instance_id,
            raise_exceptions_on_send=raise_exceptions_on_send,
            log_exceptions_if_not_raised=log_exceptions_if_not_raised,
        )

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        Upload instance resource to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials | None
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        log_exceptions_if_not_raised : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.

        Returns
        -------

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # get execution report
        if execution_report is None:
            execution_report = get_default_execution_report()
        # update execution report
        execution_report.update(files_uploaded=1)

        try:
            return update_instance_resource(
                file_content=self.file_content,
                file_name=self.file_name,
                resource_file_id=self.resource_file_id,
                model_instance_id=self.model_instance_id,
                api_credentials=api_credentials,
            )
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class TimeValuesOutput:
    """
    This class contains time values to be sent to Oi Analytics.

    Attributes
    ----------
    output_type : 'time_values'
        Denotes the specific type of the current output object.
    data : pandas.Series | pandas.DataFrame
        Data structure containing time values.
    units : dict, optional
        Dictionary that maps ID of a time data to its corresponding unit.
    use_external_reference : bool, default False
        Whether to use external reference.
    timestamp_index_name : str, default 'timestamp'
        Index name of data containing time values.
    create_upload_event : bool, default True
        Whether to create a file upload event when uploading to Oi Analytics.

    Methods
    -------
    from_pandas(self, data, file_name, file_type):
        Initialize an instance of TimeValuesOutput with data from a pandas.Series or a pandas.DataFrame.

    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions: bool = False):
        Upload time values to Oi Analytics.
    """

    def __init__(
        self,
        data: Union[pd.Series, pd.DataFrame],
        units: Optional[dict] = None,
        rename_data: bool = True,
        use_external_reference: bool = False,
        timestamp_index_name: str = "timestamp",
        create_upload_event: Optional[bool] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """
        Initialize a new instance of TimeValuesOutput.

        Parameters
        ----------
        data : pandas.Series | pandas.DataFrame
            Data structure containing time values.
        units : dict, optional
            Dictionary that maps reference of a time data to its corresponding unit.
        rename_data : bool, default True
            Whether to rename data with its reference.
        use_external_reference : bool, default False
            Whether to use external reference.
        timestamp_index_name : str, default 'timestamp'
            Index name of data containing time values.
        create_upload_event : bool, default True
            Whether to create a file upload event when uploading to Oi Analytics.
        """
        self.output_type = "time_values"
        self.create_upload_event = create_upload_event

        # Rename data if specified
        data_df = data.to_frame() if isinstance(data, pd.Series) else data

        if rename_data is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Data can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="any", values_type="scalar", mode="reference"
            )
            self.data = data_df.rename(columns=output_dict)
        else:
            self.data = data_df

        # Specify units
        if units is None:
            model_exec = get_default_model_execution()
            if model_exec is not None:
                self.units = {
                    output_data.reference: output_data.unit.label
                    for output_data in model_exec.get_data_output_dict(
                        data_type="any", values_type="scalar", mode="object"
                    ).values()
                }
            else:
                self.units = units
        else:
            self.units = units

        self.use_external_reference = use_external_reference

        self.timestamp_index_name = timestamp_index_name

        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        create_upload_event: Optional[bool] = None,
    ) -> None:
        """
        Upload time values to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.
        create_upload_event : bool, default True
            Whether to create a file upload event when uploading to Oi Analytics.

        Returns
        -------
        None

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # get execution report
        if execution_report is None:
            execution_report = get_default_execution_report()

        if create_upload_event is None:
            if self.create_upload_event is None:
                model_exec = get_default_model_execution()
                if (
                    model_exec is not None
                    and model_exec.pythonModelInstance.dataExchangeMode
                    == "FILE_PROCESSING"
                ):
                    create_upload_event = False
                else:
                    create_upload_event = True
            else:
                create_upload_event = self.create_upload_event

        # send data
        try:
            response = api.insert_time_values(
                data=self.data,
                units=self.units,
                use_external_reference=self.use_external_reference,
                timestamp_index_name=self.timestamp_index_name,
                create_upload_event=create_upload_event,
                api_credentials=api_credentials,
            )
            # update execution report
            execution_report.update(
                time_values_updated=response.get(
                    "numberOfValuesSuccessfullyInserted", 0
                ),
                errors=len(response.get("errors", [])),
            )
            return response
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class VectorTimeValuesOutput:
    """
    This class contains vector time values to be sent to Oi Analytics.

    Attributes
    ----------
    output_type : {'time_vector_values'}
        Denotes the specific type of the current output object.
    data : pandas.Series | pandas.DataFrame
        Data structure containing time values.
    data_reference : list of str
            List of data references (one data reference for each vector data).
    values_units : dict, optional
        Dictionary that maps ID of a time vector data to its corresponding values unit.
    index_units : dict, optional
        Dictionary that maps ID of a time vector data to its corresponding indexes unit.
    use_external_reference : bool, default False
        Whether to use external reference.
    timestamp_index_name : str, default 'timestamp'
        Index name of data containing time values.
    create_upload_event : bool, default True
        Whether to create a file upload event when uploading to Oi Analytics.

    Methods
    -------
    from_pandas(self, data, file_name, file_type):
        Initialize an instance of InstanceResourceOutput with data from a pandas.Series or a pandas.DataFrame.

    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions: bool = False):
        Upload vector time values to Oi Analytics.
    """

    def __init__(
        self,
        data: List[pd.DataFrame],
        data_reference: List[str],
        rename_data: bool = True,
        values_units: Optional[dict[str, str]] = None,
        index_units: Optional[dict[str, str]] = None,
        use_external_reference: bool = False,
        timestamp_index_name: str = "timestamp",
        create_upload_event: Optional[bool] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """

        Parameters
        ----------
        data : list of pandas.DataFrame
            List of vector time values.
        data_reference : list of str
            List of data references (one data reference for each vector data).
        rename_data : bool, default True
            Whether to rename data.
        values_units : dict, optional
            Dictionary that maps vector data to its corresponding unit.
        index_units : dict, optional
            Dictionary that maps vector data index to its corresponding unit.
        use_external_reference : bool, default
            Whether to use external reference.
        timestamp_index_name : str, default 'timestamp'
            Index name of each pandas.DataFrame containing vector time values.
        create_upload_event : bool, default True
            Whether to create a file upload event when uploading to Oi Analytics.
        """
        self.output_type = "time_vector_values"
        self.create_upload_event = create_upload_event

        if rename_data is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Data can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="time", values_type="vector", mode="reference"
            )
            data_reference = [
                output_dict[source_code_name] for source_code_name in data_reference
            ]

        self.data_reference = data_reference
        self.data = data

        # Specify values units
        if values_units is None:
            model_exec = get_default_model_execution()
            if model_exec is not None:
                self.values_units = {
                    output_data.reference: output_data.valueUnit.label
                    for output_data in model_exec.get_data_output_dict(
                        data_type="time", values_type="vector", mode="object"
                    ).values()
                    if output_data.reference in data_reference
                }
            else:
                self.values_units = values_units
        else:
            self.values_units = values_units

        # Specify index unit
        if index_units is None:
            model_exec = get_default_model_execution()
            if model_exec is not None:
                self.index_units = {
                    output_data.reference: output_data.indexUnit.label
                    for output_data in model_exec.get_data_output_dict(
                        data_type="time", values_type="vector", mode="object"
                    ).values()
                    if output_data.reference in data_reference
                }
            else:
                self.index_units = index_units
        else:
            self.index_units = index_units

        self.use_external_reference = use_external_reference

        self.timestamp_index_name = timestamp_index_name

        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        create_upload_event: Optional[bool] = None,
    ) -> None:
        """
        Upload vector time values to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.
        create_upload_event : bool, default True
            Whether to create a file upload event when uploading to Oi Analytics.

        Returns
        -------
        None

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # get execution report
        if execution_report is None:
            execution_report = get_default_execution_report()

        if create_upload_event is None:
            if self.create_upload_event is None:
                model_exec = get_default_model_execution()
                if (
                    model_exec is not None
                    and model_exec.pythonModelInstance.dataExchangeMode
                    == "FILE_PROCESSING"
                ):
                    create_upload_event = False
                else:
                    create_upload_event = True
            else:
                create_upload_event = self.create_upload_event

        # send data
        try:
            response = api.insert_vector_time_values(
                data=self.data,
                data_reference=self.data_reference,
                index_units=self.index_units,
                values_units=self.values_units,
                use_external_reference=self.use_external_reference,
                timestamp_index_name=self.timestamp_index_name,
                create_upload_event=create_upload_event,
                api_credentials=api_credentials,
            )

            # update execution report
            execution_report.update(
                errors=len(response.get("errors", [])),
                time_vector_values_updated=response.get(
                    "numberOfValuesSuccessfullyInserted", 0
                ),
            )
            return response
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class BatchValuesOutput:
    """
    This class contains batch values to be sent to Oi Analytics.

    Attributes
    ----------
    output_type : {'batch_values'}
        Denotes the specific type of the current output object.
    batch_type_id : str
        Unique identifier of batch type.
    data : pandas.Series | pandas.DataFrame
        Data structure containing time values.
    units : dict, optional
        Dictionary that maps ID of a time data to its corresponding unit.
    batch_id_index_name : str, default 'batch_id'
        Index name of data containing time values.

    Methods
    -------
    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions: bool = False):
        Upload batch values to Oi Analytics.
    """

    def __init__(
        self,
        batch_type_id: str,
        data: Union[pd.Series, pd.DataFrame],
        units: Optional[dict] = None,
        batch_id_index_name: str = "batch_id",
        rename_data: bool = True,
        create_upload_event: Optional[bool] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """

        Parameters
        ----------
        batch_type_id : str
            Unique identifier of the batch type.
        data : pandas.Series | pandas.DataFrame
            Batch values.
        units : dict, Optional
            Dictionary that maps vector data to its corresponding unit.
        batch_id_index_name : str, default 'batch_id'
            Index name of data containing batch values.
        rename_data : bool, default True
            Whether to rename with technical value (value id).
        create_upload_event : bool, default True
            Whether to create a file upload event when uploading to Oi Analytics.
        """
        self.output_type = "batch_values"
        self.batch_type_id = batch_type_id
        self.create_upload_event = create_upload_event

        # Rename data if specified
        data_df = data.to_frame() if isinstance(data, pd.Series) else data

        if rename_data is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Data can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="batch", values_type="scalar", mode="id"
            )
            self.data = data_df.rename(columns=output_dict)
        else:
            self.data = data_df

        # Specify units
        if units is None:
            model_exec = get_default_model_execution()
            if model_exec is not None:
                self.units = {
                    output_data.id: output_data.unit.id
                    for output_data in model_exec.get_data_output_dict(
                        data_type="batch", values_type="scalar", mode="object"
                    ).values()
                }
            else:
                self.units = None
        else:
            self.units = units

        self.batch_id_index_name = batch_id_index_name

        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        create_upload_event: Optional[bool] = None,
    ) -> None:
        """
        Upload batch values to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.
        create_upload_event : bool, default True
            Whether to create a file upload event when uploading to Oi Analytics.

        Returns
        -------
        None

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # get execution report
        if execution_report is None:
            execution_report = get_default_execution_report()

        if create_upload_event is None:
            if self.create_upload_event is None:
                model_exec = get_default_model_execution()
                if (
                    model_exec is not None
                    and model_exec.pythonModelInstance.dataExchangeMode
                    == "FILE_PROCESSING"
                ):
                    create_upload_event = False
                else:
                    create_upload_event = True
            else:
                create_upload_event = self.create_upload_event

        try:
            # send data
            response = api.update_batch_features_and_values(
                batch_type_id=self.batch_type_id,
                data=self.data,
                unit_ids=self.units,
                batch_id_index_name=self.batch_id_index_name,
                create_upload_event=create_upload_event,
                api_credentials=api_credentials,
            )

            # update execution report
            execution_report.update(batch_values_updated=int(self.data.count().sum()))
            return response
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class VectorBatchValuesOutput:
    """
    This class contains vector batch values to be sent to Oi Analytics.

    Attributes
    ----------
    output_type : {'batch_vector_values'}
        Denotes the specific type of the current output object.
    create_upload_event : bool, default True
        Whether to create a file upload event when uploading to Oi Analytics.
    data_reference : list of str
        The unique data references of the data being inserted.
    data : pandas.Series | pandas.DataFrame
        Data structure containing time values.
    values_units : dict, optional
        Dictionary that maps ID of a batch vector data to its corresponding value unit.
    index_units : dict, optional
        Dictionary that maps ID of a batch vector data to its corresponding index unit.
    batch_id_index_name : str, default 'batch_id'
        Index name of data containing time values.

    Methods
    -------
    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions: bool = False):
        Upload vector batch values to OI Analytics.
    """

    def __init__(
        self,
        data: List[pd.DataFrame],
        data_reference: List[str],
        values_units: Optional[dict[str, str]] = None,
        index_units: Optional[dict[str, str]] = None,
        batch_id_index_name: str = "batch_id",
        rename_data: bool = True,
        create_upload_event: Optional[bool] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """

        Parameters
        ----------
        data : list of data
            List of batch vector values.
        data_reference : list of str
            List of data references.
        values_units : dict, optional
            Dictionary that maps reference of a vector data to its corresponding unit.
        index_units : dict, None
            Dictionary that maps reference of a vector data to its corresponding index unit.
        batch_id_index_name : str, default 'batch_id'
            Index name of data containing batch values.
        rename_data : bool, default True
            Whether to rename with technical value (value id).
        create_upload_event : bool, default True
            Whether to create a file upload event when uploading to Oi Analytics.
        """
        self.output_type = "batch_vector_values"
        self.create_upload_event = create_upload_event

        if rename_data is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Data can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="batch", values_type="vector", mode="reference"
            )
            data_reference = [
                output_dict[source_code_name] for source_code_name in data_reference
            ]

        self.data_reference = data_reference
        self.data = data

        # Specify units
        if values_units is None:
            model_exec = get_default_model_execution()
            if model_exec is not None:
                self.values_units = {
                    output_data.reference: output_data.valueUnit.label
                    for output_data in model_exec.get_data_output_dict(
                        data_type="batch", values_type="vector", mode="object"
                    ).values()
                    if output_data.reference in data_reference
                }
        else:
            self.values_units = values_units

        # Specify units
        if index_units is None:
            model_exec = get_default_model_execution()
            if model_exec is not None:
                self.index_units = {
                    output_data.reference: output_data.indexUnit.label
                    for output_data in model_exec.get_data_output_dict(
                        data_type="batch", values_type="vector", mode="object"
                    ).values()
                    if output_data.reference in data_reference
                }
        else:
            self.index_units = index_units

        self.batch_id_index_name = batch_id_index_name

        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        create_upload_event: Optional[bool] = None,
    ) -> None:
        """
        Upload vector batch values to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.
        create_upload_event : bool, default True
            Whether to create a file upload event when uploading to Oi Analytics.

        Returns
        -------
        None

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # get execution report
        if execution_report is None:
            execution_report = get_default_execution_report()

        if create_upload_event is None:
            if self.create_upload_event is None:
                model_exec = get_default_model_execution()
                if (
                    model_exec is not None
                    and model_exec.pythonModelInstance.dataExchangeMode
                    == "FILE_PROCESSING"
                ):
                    create_upload_event = False
                else:
                    create_upload_event = True
            else:
                create_upload_event = self.create_upload_event

        # send data
        try:
            response = api.update_vector_batch_values(
                data=self.data,
                data_reference=self.data_reference,
                index_units=self.index_units,
                values_units=self.values_units,
                batch_id_index_name=self.batch_id_index_name,
                create_upload_event=create_upload_event,
                api_credentials=api_credentials,
            )

            # update execution report
            execution_report.update(
                errors=len(response.get("errors", [])),
                batch_vector_values_updated=response.get(
                    "numberOfValuesSuccessfullyInserted", 0
                ),
            )
            return response
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class BatchFeaturesOutput:
    """
    This class contains batch features to be sent to Oi Analytics.

    Attributes
    ----------
    output_type : {'batch_features'}
        Indicates the output type to be sent to Oi Analytics.
    batch_type_id : str
        Unique identifier of batch type.
    data : pandas.Series | pandas.DataFrame
        Data structure containing time values.
    batch_id_index_name : str, default 'batch_id'
        Index name of data containing batch features.
    create_upload_event : bool, optional
        Whether to create a file upload event when uploading to Oi Analytics.

    Methods
    -------
    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions: bool = False):
        Upload batch features to OI Analytics.
    """

    def __init__(
        self,
        batch_type_id: str,
        data: Union[pd.Series, pd.DataFrame],
        rename_features: bool = True,
        batch_id_index_name: str = "batch_id",
        create_upload_event: Optional[bool] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """
        Initialize a new instance of BatchFeaturesOutput.

        Parameters
        ----------
        batch_type_id : str
            Unique identifier of the batch type.
        data : pandas.Series | pandas.DataFrame
            Data containing batch features output.
        rename_features : bool, default True
            Whether to rename features from source code names to data IDs.
        batch_id_index_name : str, default 'batch_id'
            Index name of data containing batch values.
        """
        self.output_type = "batch_features"
        self.batch_type_id = batch_type_id
        self.create_upload_event = create_upload_event

        # Rename data if specified
        data_df = data.to_frame() if isinstance(data, pd.Series) else data

        if rename_features is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Features can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="batch", values_type="categorical", mode="id"
            )
            self.data = data_df.rename(columns=output_dict)
        else:
            self.data = data_df

        self.batch_id_index_name = batch_id_index_name

        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        create_upload_event: Optional[bool] = None,
    ) -> None:
        """
        Upload batch features to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.
        create_upload_event : bool, default True
            Whether to create a file upload event when uploading to Oi Analytics.

        Returns
        -------

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # update execution report
        if execution_report is None:
            execution_report = get_default_execution_report()

        if create_upload_event is None:
            if self.create_upload_event is None:
                model_exec = get_default_model_execution()
                if (
                    model_exec is not None
                    and model_exec.pythonModelInstance.dataExchangeMode
                    == "FILE_PROCESSING"
                ):
                    create_upload_event = False
                else:
                    create_upload_event = True
            else:
                create_upload_event = self.create_upload_event

        # send data
        try:
            response = api.update_batch_features_and_values(
                batch_type_id=self.batch_type_id,
                data=self.data,
                feature_columns=list(self.data.columns),
                batch_id_index_name=self.batch_id_index_name,
                create_upload_event=create_upload_event,
                api_credentials=api_credentials,
            )

            # update execution report
            execution_report.update(batch_tags_updated=int(self.data.count().sum()))
            return response
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class BatchesOutput:
    """
    This class contains batches to be sent to Oi Analytics.

    Attributes
    ----------
    output_type : {'batches'}
        Denotes the specific type of the current output object.
    batch_type_id : str
        Unique identifier of batch type.
    steps : pandas.DataFrame
        The steps of the batches to insert. The DataFrame is indexed by batch name (named after batch_name_index_name).
            The content of this DataFrame must be:
            - The step ID (named after step_id_index_name)
            - The start and end dates of the steps (named after start_date_name and end_date_name)
            - Optional asset or tag localisation columns
    on_duplicates_keep : {'first', 'last', False}, default 'last'
        Indicate which DataFrame row to keep in case of duplicates (Cf. pd.DataFrame.drop_duplicates).
    batch_name_index_name : str, default 'batch_name'
        The name of the index level containing the batch name in the DataFrames.
    step_id_index_name : str, default 'step_id_index_name'
        The name of the index level (or column) containing the step id in the steps DataFrame.
    start_date_name : str, default 'start'
        The name of the column containing the start date of the steps id in the steps DataFrame.
    end_date_name : str, default 'end'
        The name of the column containing the end date of the steps id in the steps DataFrame.
    asset_localisation_column : str, optional
        The name of the column containing the asset IDs allowing the localisation of steps.
    tag_localisation_columns : str | list of str, optional
        The name of the columns containing the tag value IDs allowing the localisation of steps.
    vector_data_values : pandas.DataFrame | list of pandas.DataFrame, optional
        The vector data values to insert or update. Must be indexed by batch name (Cf. batch_name_index_name).
        Each column is a vector index.
    values : pandas.DataFrame, optional
        The data values to update or insert. The DataFrame must be indexed by batch name (Cf. batch_name_index_name).
        Each column is named after the data ID.
    values_unit_ids : dict, optional
        Dictionary that maps data id to its corresponding unit id.
    features : pandas.DataFrame, optional
        The feature values to update or insert. The DataFrame must be indexed by batch name (Cf. batch_name_index_name).
        Each column is named after the feature ID.
    vector_data_references : list of str, optional
        The list of the vector data references matching the list  of vector data values DataFrames.
    vector_data_index_units : dict, None
        Dictionary that maps reference of a vector data to its corresponding index unit.
    vector_data_values_units : dict, None
        Dictionary that maps reference of a vector data to its corresponding unit.

    Methods
    -------
    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
        Upload batches to OI Analytics.
    """

    def __init__(
        self,
        batch_type_id: str,
        steps: pd.DataFrame,
        values: Optional[pd.DataFrame] = None,
        values_unit_ids: Optional[dict] = None,
        rename_data: bool = True,
        features: Optional[pd.DataFrame] = None,
        rename_features: bool = True,
        vector_data_values: Optional[Union[pd.DataFrame, List[pd.DataFrame]]] = None,
        vector_data_references: Optional[List[str]] = None,
        vector_data_index_units: Optional[dict] = None,
        vector_data_values_units: Optional[dict] = None,
        rename_vector_data: bool = True,
        on_duplicates_keep: str = "last",
        batch_name_index_name: str = "batch_name",
        step_id_index_name: str = "step_id",
        start_date_name: str = "start",
        end_date_name: str = "end",
        asset_localisation_column: Optional[str] = None,
        tag_localisation_columns: Optional[Union[str, List[str]]] = None,
        create_upload_event: Optional[bool] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """
        Initialize a new instance of BatchesOutput.

        Parameters
        ----------
        batch_type_id : str
            Unique identifier of the batch type.
        steps : pandas.DataFrame
            The steps of the batches to insert. The DataFrame is indexed by batch name (named after batch_name_index_name).
            The content of this DataFrame must be:
            - The step ID (named after step_id_index_name)
            - The start and end dates of the steps (named after start_date_name and end_date_name)
            - Optional asset or tag localisation columns
        values : pandas.DataFrame, optional
            The data values to update or insert. The DataFrame must be indexed by batch name (Cf. batch_name_index_name).
            Each column is named after the data ID.
        values_unit_ids : dict, optional
            Dictionary that maps data id to its corresponding unit id.
        rename_data : bool, default True
            Whether to rename data
        features : pandas.DataFrame, optional
            The feature values to update or insert. The DataFrame must be indexed by batch name (Cf. batch_name_index_name).
            Each column is named after the feature ID.
        rename_features : bool, default True
            Whether to rename features with IDs.
        vector_data_values : pandas.DataFrame | list of pandas.DataFrame, optional
            The vector data values to insert or update. Must be indexed by batch name (Cf. batch_name_index_name).
            Each column is a vector index.
        vector_data_references : list of str, optional
            The list of the vector data references matching the list  of vector data values DataFrames.
        vector_data_index_units : dict, None
            Dictionary that maps reference of a vector data to its corresponding index unit.
        vector_data_values_units : dict, None
            Dictionary that maps reference of a vector data to its corresponding unit.
        rename_vector_data : bool, default True
            Whether to rename the vector data.
        on_duplicates_keep : {'first', 'last', False}, default 'last'
            Indicate which DataFrame row to keep in case of duplicates (Cf. pd.DataFrame.drop_duplicates).
        batch_name_index_name : str, default 'batch_name'
            The name of the index level containing the batch name in the DataFrames.
        step_id_index_name : str, default 'step_id_index_name'
            The name of the index level (or column) containing the step id in the steps DataFrame.
        start_date_name : str, default 'start'
            The name of the column containing the start date of the steps id in the steps DataFrame.
        end_date_name : str, default 'end'
            The name of the column containing the end date of the steps id in the steps DataFrame.
        asset_localisation_column : str, optional
            The name of the column containing the asset IDs allowing the localisation of steps.
        tag_localisation_columns : str | list of str, optional
            The name of the columns containing the tag value IDs allowing the localisation of steps.
        """
        self.output_type = "batches"
        self.batch_type_id = batch_type_id
        self.steps = steps
        self.on_duplicates_keep = on_duplicates_keep
        self.batch_name_index_name = batch_name_index_name
        self.step_id_index_name = step_id_index_name
        self.start_date_name = start_date_name
        self.end_date_name = end_date_name
        self.asset_localisation_column = asset_localisation_column
        self.tag_localisation_columns = tag_localisation_columns
        self.vector_data_values = vector_data_values
        self.create_upload_event = create_upload_event

        # Rename data if specified
        if rename_data is True and values is not None:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Data can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="batch", values_type="scalar", mode="id"
            )

            # Rename values DataFrame
            self.values = values.rename(columns=output_dict)

            # Rename data in units dict
            if values_unit_ids is not None:
                self.values_unit_ids = {
                    output_dict[k]: v for k, v in values_unit_ids.items()
                }
            else:
                self.values_unit_ids = None
        else:
            self.values = values
            self.values_unit_ids = values_unit_ids

        # Rename features if specified
        if rename_features is True and features is not None:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Features can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="batch", values_type="categorical", mode="id"
            )

            # Rename values DataFrame
            self.features = features.rename(columns=output_dict)
        else:
            self.features = features

        # Rename vector data if specified
        if rename_vector_data is True and vector_data_values is not None:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Vector data can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="batch", values_type="vector", mode="reference"
            )
            self.vector_data_references = [
                output_dict[source_code_name]
                for source_code_name in vector_data_references
            ]

            # Rename data in index units dict
            if vector_data_index_units is not None:
                self.vector_data_index_units = {
                    output_dict[k]: v for k, v in vector_data_index_units.items()
                }
            else:
                self.vector_data_index_units = None

            # Rename data in values units dict
            if vector_data_values_units is not None:
                self.vector_data_values_units = {
                    output_dict[k]: v for k, v in vector_data_values_units.items()
                }
            else:
                self.vector_data_values_units = None
        else:
            self.vector_data_references = vector_data_references
            self.vector_data_index_units = None
            self.vector_data_values_units = None

        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        create_upload_event: Optional[bool] = None,
    ) -> None:
        """
        Upload batches to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.
        create_upload_event : bool, optional
            Whether to create a file upload event when uploading to Oi Analytics.

        Methods
        -------
        send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
            Upload batches to OI Analytics.

        Returns
        -------
        None

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # update execution report
        if execution_report is None:
            execution_report = get_default_execution_report()

        if create_upload_event is None:
            if self.create_upload_event is None:
                model_exec = get_default_model_execution()
                if (
                    model_exec is not None
                    and model_exec.pythonModelInstance.dataExchangeMode
                    == "FILE_PROCESSING"
                ):
                    create_upload_event = False
                else:
                    create_upload_event = True
            else:
                create_upload_event = self.create_upload_event

        # send data
        try:
            response = api.create_or_update_batches(
                batch_type_id=self.batch_type_id,
                steps=self.steps,
                values=self.values,
                values_unit_ids=self.values_unit_ids,
                features=self.features,
                vector_data_values=self.vector_data_values,
                vector_data_references=self.vector_data_references,
                vector_data_index_units=self.vector_data_index_units,
                vector_data_values_units=self.vector_data_values_units,
                on_duplicates_keep=self.on_duplicates_keep,
                batch_name_index_name=self.batch_name_index_name,
                step_id_index_name=self.step_id_index_name,
                start_date_name=self.start_date_name,
                end_date_name=self.end_date_name,
                asset_localisation_column=self.asset_localisation_column,
                tag_localisation_columns=self.tag_localisation_columns,
                create_upload_event=create_upload_event,
                api_credentials=api_credentials,
            )

            # update execution report
            execution_report.update(batch_created_updated=len(response))
            return response
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class EventsOutput:
    """
    This class contains events to be sent to Oi Analytics.

    Attributes
    ----------
    output_type : {'events'}
        Denotes the specific type of the current output object.
    event_type_id : str
            Unique identifier of the event type.
        events : pandas.DataFrame
            The events to create or update. The DataFrame is indexed by event id (left empty is the event si to be created).
            The content of this DataFrame must be:
            - The start and end dates of the events (named after start_date_name and end_date_name)
            - A column for each event data pr feature
            - An optional column containing the list of assets events are attached to
        value_columns: List[str], default []
            The list of column names containing event values
        feature_columns: List[str], default []
            The list of column names containing event features
        values_unit_ids : dict, optional
            Dictionary that maps data id to its corresponding unit id.
        rename_data : bool, default True
            Whether to rename data
        rename_features : bool, default True
            Whether to rename features with IDs.
        start_date_name : str, default 'start'
            The name of the column containing the start date of the events
        end_date_name : str, default 'end'
            The name of the column containing the end date of the events
        asset_column : str, optional
            The name of the column containing the list of asset IDs events are attached to

    Methods
    -------
    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
        Upload events to OI Analytics.
    """

    def __init__(
        self,
        event_type_id: str,
        events: pd.DataFrame,
        value_columns: List[str] = [],
        feature_columns: List[str] = [],
        asset_column: Optional[List[dict]] = None,
        values_unit_ids: Optional[dict] = None,
        start_date_name: str = "start",
        end_date_name: str = "end",
        description_name: str = "description",
        create_upload_event: Optional[bool] = None,
        rename_data: bool = True,
        rename_features: bool = True,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """
        Initialize a new instance of EventsOutput.

        Parameters
        ----------
        event_type_id : str
            Unique identifier of the event type.
        events : pandas.DataFrame
            The events to create or update. The DataFrame is indexed by event id (left empty is the event si to be created).
            The content of this DataFrame must be:
            - The start and end dates of the events (named after start_date_name and end_date_name)
            - A column for each event data pr feature
            - An optional column containing the list of assets events are attached to
        value_columns: List[str], default []
            The list of column names containing event values
        feature_columns: List[str], default []
            The list of column names containing event features
        values_unit_ids : dict, optional
            Dictionary that maps data id to its corresponding unit id.
        rename_data : bool, default True
            Whether to rename data
        rename_features : bool, default True
            Whether to rename features with IDs.
        start_date_name : str, default 'start'
            The name of the column containing the start date of the events
        end_date_name : str, default 'end'
            The name of the column containing the end date of the events
        asset_column : str, optional
            The name of the column containing the list of asset IDs events are attached to
        """

        self.output_type = "events"
        self.event_type_id = event_type_id
        self.events = events
        self.value_columns = value_columns
        self.feature_columns = feature_columns
        self.asset_column = asset_column
        self.start_date_name = start_date_name
        self.end_date_name = end_date_name
        self.description_name = description_name
        self.create_upload_event = create_upload_event
        self.values_unit_ids = values_unit_ids

        # Rename data if specified
        if rename_data is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Data can't be renamed without a current model execution context set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="event", values_type="scalar", mode="id"
            )

            # Rename values DataFrame
            self.events = self.events.rename(columns=output_dict)

            # Rename data in units dict
            if values_unit_ids is not None:
                self.values_unit_ids = {
                    output_dict[k]: v for k, v in self.values_unit_ids.items()
                }
            else:
                self.values_unit_ids = None

        # Rename features if specified
        if rename_features is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Features can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="event", values_type="categorical", mode="id"
            )

            # Rename values DataFrame
            self.events = self.events.rename(columns=output_dict)

        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        create_upload_event: Optional[bool] = None,
    ) -> None:
        """
        Upload events to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.
        create_upload_event : bool, optional
            Whether to create a file upload event when uploading to Oi Analytics.

        Methods
        -------
        send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
            Upload batches to OI Analytics.

        Returns
        -------
        None

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # update execution report
        if execution_report is None:
            execution_report = get_default_execution_report()

        if create_upload_event is None:
            if self.create_upload_event is None:
                model_exec = get_default_model_execution()
                if (
                    model_exec is not None
                    and model_exec.pythonModelInstance.dataExchangeMode
                    == "FILE_PROCESSING"
                ):
                    create_upload_event = False
                else:
                    create_upload_event = True
            else:
                create_upload_event = self.create_upload_event

        # send data
        try:
            response = api.create_or_update_events(
                event_type_id=self.event_type_id,
                events=self.events,
                value_columns=self.value_columns,
                feature_columns=self.feature_columns,
                asset_column=self.asset_column,
                values_unit_ids=self.values_unit_ids,
                start_date_name=self.start_date_name,
                end_date_name=self.end_date_name,
                description_name=self.description_name,
                create_upload_event=create_upload_event,
                api_credentials=api_credentials,
            )

            # update execution report
            execution_report.update(events_created_updated=len(response))
            return response
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class Delay:
    """
    This class contains time delays to be added to the execution of a Python model.

    Attributes
    ----------
    output_type : {'delay'}
        Denotes the specific type of the current output object.
    duration : int
        Time delay in the execution, in seconds.

    Methods
    -------
    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
        Add time delays to Python model execution in OI Analytics.
    """

    def __init__(self, duration=5):
        """
        Initialize a new instance of Delay.

        Parameters
        ----------
        duration : int
            Time delay in the execution, in seconds.
        """
        self.output_type = "delay"
        self.duration = duration

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        **kwargs,
    ):
        """
        Add time delays to Python model execution in OI Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials | None
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.

        Returns
        -------

        """
        # update execution report
        if execution_report is None:
            execution_report = get_default_execution_report()

        time.sleep(self.duration)


class CustomTextOutput:
    """
    This class contains customized texts to be added to the model execution report.

    Attributes
    ----------
    type : {'text'}
        Denotes the specific type of the current output object.
    content : str
        Customized text to be added to the model execution report.

    Methods
    -------
    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
        Add the customized text to the Python model execution report.
    """

    def __init__(
        self,
        content: str,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """
        Initialize a new instance of CustomTextOutput.

        Parameters
        ----------
        content : str
        """
        self.type = "text"
        self.content = content
        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        Add the customized text to the Python model execution report.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials | None
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport | None
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.

        Returns
        -------
        None

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # Get the default execution report if not provided
        if execution_report is None:
            execution_report = get_default_execution_report()

        # Update the execution report
        try:
            execution_report.customOutput = CustomModelOutput(
                type=self.type, content=self.content
            )
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class CustomJsonOutput:
    """
    This class contains customized json output to be added to the model execution report.

    Attributes
    ----------
    type : {'json'}
        Denotes the specific type of the current output object.
    content : list | str | dict
        Customized content to send added to the model execution report.

    Methods
    -------
    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
        Add customized content to model execution report.
    """

    def __init__(
        self,
        content,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """

        Parameters
        ----------
        content : list | str | dict
        """
        self.type = "json"
        self.content = content
        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        **kwargs,
    ):
        """

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.

        Returns
        -------

        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # Get the default execution report if not provided
        if execution_report is None:
            execution_report = get_default_execution_report()

        # Update the execution report
        try:
            execution_report.customOutput = CustomModelOutput(
                type=self.type, content=self.content
            )
        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class BatchComputationJob:
    """
    This class contains a command to create a batch computation job in OIAnalytics.

    Attributes
    ----------
    output_type : {'batch_computation_job'}
        Indicates the output type to be sent to OIAnalytics
    batch_type_id : str
        The batch type ID attached to batches to be computed
    data_ids : list of str, optional
        List of data IDs to be computed. If empty, all batch data are computed
    batch_ids : list of str, optional
        List of batch IDs to be computed. If empty, all batches are computed
    """

    def __init__(
        self,
        batch_type_id: str,
        data_ids: Optional[List[str]] = None,
        batch_ids: Optional[List[str]] = None,
        raise_exceptions_on_send: bool = True,
        log_exceptions_if_not_raised: bool = True,
    ):
        """
        Initialize a new instance of BatchComputationJob

        Attributes
        ----------
        output_type : {'batch_computation_job'}
            Indicates the output type to be sent to OIAnalytics
        batch_type_id : str
            The batch type ID attached to batches to be computed
        data_ids : list of str, optional
            List of data IDs to be computed. If None, all batch data are computed
        batch_ids : list of str, optional
            List of batch IDs to be computed. If None, all batches are computed
        """

        self.output_type = "batch_computation_job"
        self.batch_type_id = batch_type_id
        self.data_ids = data_ids
        self.batch_ids = batch_ids
        self.raise_exceptions_on_send = raise_exceptions_on_send
        self.log_exceptions_if_not_raised = log_exceptions_if_not_raised

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        Send the batch computation job creation command to OIAnalytics

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used
        execution_report : ExecutionReport, optional
            Execution report for external use
        print_exceptions : bool, default True
            Whether to print exception
        raise_exceptions : bool, default False
            Whether to raise exception

        Returns
        -------
        dict
            ID and computation job type, with the keys 'id' and 'type'
        """

        # Init optional args
        if raise_exceptions is None:
            raise_exceptions = self.raise_exceptions_on_send
        if log_exceptions_if_not_raised is None:
            log_exceptions_if_not_raised = self.log_exceptions_if_not_raised

        # update execution report
        if execution_report is None:
            execution_report = get_default_execution_report()

        # send data
        try:
            response = api.endpoints.computation_jobs.create_batch_computation_jobs(
                batch_type_id=self.batch_type_id,
                data_ids=self.data_ids,
                batch_ids=self.batch_ids,
                api_credentials=api_credentials,
            )

            execution_report.update(batch_created_updated=len(self.batch_ids))

            return response

        except Exception:
            execution_report.update(errors=1)
            if raise_exceptions is True:
                raise
            if log_exceptions_if_not_raised is True:
                print(
                    f"Error when trying to send to OIAnalytics:\n{traceback.format_exc()}"
                )


class OIModelOutputs:
    """
    This class contains all the outputs objects to be sent to Oi Analytics.

    Attributes
    ----------
    output_type : {'outputs_queue'}
        Denotes the specific type of the current output object.
    model_outputs : list of objects
        List of output objects resulted from the model.

    Methods
    -------
    send_to_oianalytics(self, api_credentials, execution_report, print_exceptions, raise_exceptions):
        Upload all model outputs to Oi Analytics.
    """

    def __init__(self):
        self.output_type = "outputs_queue"
        self.model_outputs = []

    def add_output(
        self,
        output_object: Union[
            FileOutput,
            TimeValuesOutput,
            BatchValuesOutput,
            VectorTimeValuesOutput,
            VectorBatchValuesOutput,
            Delay,
            CustomTextOutput,
            CustomJsonOutput,
        ],
    ) -> None:
        """

        Parameters
        ----------
        output_object : FileOutput | TimeValuesOutput | BatchValuesOutput | VectorTimeValuesOutput | VectorBatchValuesOutput | Delay | CustomTextOutput | CustomJsonOutput

        Returns
        -------
        None

        """
        self.model_outputs.append(output_object)

    def send_to_oianalytics(
        self,
        api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
        execution_report: Optional[ExecutionReport] = None,
        raise_exceptions: Optional[bool] = None,
        log_exceptions_if_not_raised: Optional[bool] = None,
        create_upload_event: Optional[bool] = None,
    ) -> None:
        """
        Upload all model outputs to Oi Analytics.

        Parameters
        ----------
        api_credentials : api.OIAnalyticsAPICredentials, optional
            The credentials to use to query the API. If None, previously set default credentials are used.
        execution_report : ExecutionReport, optional
            Execution report for external use.
        print_exceptions : bool, default True
            Whether to print exception.
        raise_exceptions : bool, default False
            Whether to raise exception.
        create_upload_event : bool, optional
            Whether to create a file upload event when uploading to Oi Analytics.

        Returns
        -------
        None

        """
        for model_output in self.model_outputs:
            model_output.send_to_oianalytics(
                api_credentials=api_credentials,
                execution_report=execution_report,
                raise_exceptions=raise_exceptions,
                log_exceptions_if_not_raised=log_exceptions_if_not_raised,
                create_upload_event=create_upload_event,
            )



from typing import Optional
import requests

from oianalytics.api import _credentials


__all__ = [
    "get_tag_keys",
    "get_single_tag_key",
    "delete_single_tag_key",
    "get_tag_values",
    "get_single_tag_value",
    "delete_single_tag_value",
    "create_tag_value",
    "update_tag_value",
    "create_tag_key",
    "update_tag_key",
]


def get_tag_keys(
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_single_tag_key(
    tag_key_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys/{tag_key_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def delete_single_tag_key(
    tag_key_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Delete an existing tag key with a given ID.

    Parameters
    ----------
    tag_key_id : str
        The OIAnalytics ID of the tag key.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.

    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys/{tag_key_id}"
    response = requests.delete(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def get_tag_values(
    tag_key_id: str,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys/{tag_key_id}/values"
    response = requests.get(
        url=url,
        params={"page": page, "size": page_size},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_single_tag_value(
    tag_key_id: str,
    tag_value_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys/{tag_key_id}/values/{tag_value_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def delete_single_tag_value(
    tag_key_id: str,
    tag_value_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Delete an existing tag value with a given id.

    Parameters
    ----------
    tag_key_id : str
        The OIAnalytics ID of the tag key.
    tag_value_id : str
        The OIAnalytics ID of the tag value.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys/{tag_key_id}/values/{tag_value_id}"
    response = requests.delete(
        url=url,
        **api_credentials.auth_kwargs,
    )

    response.raise_for_status()
    return response.status_code


def create_tag_value(
    tag_key_id: str,
    value: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys/{tag_key_id}/values"
    response = requests.post(
        url=url,
        json={"value": value},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_tag_value(
    tag_key_id: str,
    tag_value_id: str,
    value: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing tag value.

    Parameters
    ----------
    tag_key_id : str
        The OIAnalytics ID of the tag key of the tag value to be updated.
    tag_value_id : str
        The OIAnalytics ID of the tag value to be updated.
    value : str
        The new value of the tag value.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys/{tag_key_id}/values/{tag_value_id}"
    response = requests.put(
        url=url,
        json={"value": value},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def create_tag_key(
    key: str,
    position: int,
    used_for_access_control: bool,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys"
    response = requests.post(
        url=url,
        json={
            "key": key,
            "position": position,
            "usedForAccessControl": used_for_access_control,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_tag_key(
    tag_key_id: str,
    key: str,
    position: int,
    used_for_access_control: bool,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing tag key based on a command.

    Parameters
    ----------
    tag_key_id : str
    key : str
        The key of the tag key.
    position : int
        The position of the tag key.
    used_for_access_control : bool
        Whether this tag key is used to control access objects and data.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/tag-keys/{tag_key_id}"
    response = requests.put(
        url=url,
        json={
            "key": key,
            "position": position,
            "usedForAccessControl": used_for_access_control,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code



from typing import Optional, List

import requests

from .. import _credentials


__all__ = ["get_users", "get_user_details", "delete_user", "create_user", "update_user"]


def get_users(
    page: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/users"
    response = requests.get(
        url=url,
        params={"page": page},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_user_details(
    user_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/users/{user_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def delete_user(
    user_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Delete an existing user with a given id.

    Parameters
    ----------
    user_id : str
        The OIAnalytics ID of the user to delete.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/users/{user_id}"
    response = requests.delete(
        url=url,
        **api_credentials.auth_kwargs,
    )

    response.raise_for_status()
    return response.status_code


def create_user(
    login: str,
    first_name: str,
    last_name: str,
    email: Optional[str],
    phone_number: Optional[str],
    language: Optional[str],
    timezone: Optional[str],
    login_allowed: Optional[bool],
    technical_user: Optional[bool],
    using_sso: Optional[bool],
    session_timeout: Optional[int],
    expiration_date: int,
    profile_id: str,
    tag_access_list: List[dict],
    main_dashboard_id: Optional[str] = None,
    main_notebook_id: Optional[str] = None,
    main_input_form_id: Optional[str] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/users"
    response = requests.post(
        url=url,
        json={
            "login": login,
            "firstName": first_name,
            "lastName": last_name,
            "email": email,
            "phoneNumber": phone_number,
            "language": language,
            "timezone": timezone,
            "loginAllowed": login_allowed,
            "technicalUser": technical_user,
            "usingSSO": using_sso,
            "sessionTimeoutInMinutes": session_timeout,
            "expirationDate": expiration_date,
            "profileId": profile_id,
            "tagAccessList": tag_access_list,
            "mainDashboardId": main_dashboard_id,
            "mainNotebookId": main_notebook_id,
            "mainInputFormId": main_input_form_id,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_user(
    user_id: str,
    login: str,
    first_name: str,
    last_name: str,
    email: str,
    phone_number: Optional[str],
    language: Optional[str],
    timezone: Optional[str],
    login_allowed: Optional[bool],
    technical_user: Optional[bool],
    using_sso: Optional[bool],
    session_timeout: Optional[int],
    expiration_date: int,
    profile_id: str,
    tag_access_list: List[dict],
    main_dashboard_id: Optional[str] = None,
    main_notebook_id: Optional[str] = None,
    main_input_form_id: Optional[str] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing user.

    Parameters
    ----------
    user_id : str
        The OIAnalytics ID of the user.
    login : str
        The login of the user.
    first_name : str
        The first name of the user.
    last_name : str
        The last name of the user.
    email : str
        The email of the user.
    phone_number : str, optional
        The phone number of the user.
    language : str
        The display language of the user.
    timezone : str
        The display language of the user.
    login_allowed : bool
        Whether the user is allowed to log in the application.
    technical_user : bool
        Whether the user is technical.
    using_sso : bool
        Whether the user is using sso to authenticate to the application.
    session_timeout : int, optional
        The session timeout, in minutes. Must be at least 1440.
        If None, a default timeout will be used.
    expiration_date : int
        The expiration date of the user. After this date, the user can no longer log in.
    profile_id : str
        The OIAnalytics ID of the user’s profile.
    tag_access_list : list of dict
        The tag access used of the user.
        It is a list of dict, where each dict contains the following keys:
        - 'tagKeyId': str, the tag key ID of the tag access
        - 'allTagValuesAccessible': bool, whether the user has access to all the tag values of the tag key
        - 'accessibleTagValueIds': list of str, the tag value IDs of the tag access.

    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.
    main_dashboard_id : str, optional
        The main of the dashboard of the user.
    main_notebook_id : str, optional
        The main of the notebook of the user.
    main_input_form_id : str, optional
        The main of the input form of the user.

    Returns
    -------
    int
        Status Code.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/users/{user_id}"
    response = requests.put(
        url=url,
        json={
            "login": login,
            "firstName": first_name,
            "lastName": last_name,
            "email": email,
            "phoneNumber": phone_number,
            "language": language,
            "timezone": timezone,
            "loginAllowed": login_allowed,
            "technicalUser": technical_user,
            "usingSSO": using_sso,
            "sessionTimeoutInMinutes": session_timeout,
            "expirationDate": expiration_date,
            "profileId": profile_id,
            "tagAccessList": tag_access_list,
            "mainDashboardId": main_dashboard_id,
            "mainNotebookId": main_notebook_id,
            "mainInputFormId": main_input_form_id,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code



from typing import Optional

import pandas as pd

from ... import api
from .. import _dtos
from .. import utils

__all__ = ["get_single_obs_time_data", "get_single_obs_batch_data"]


def get_single_obs_time_data(
    model_execution: Optional[_dtos.ModelExecution] = None,
    rename_data_to_source_code_name: bool = True,
    api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    # Get model execution from environment if not specified
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()
        if model_execution.pythonModelInstance.dataExchangeMode != "SINGLE_OBSERVATION":
            raise ValueError("The execution has to be in single observation mode")
        if model_execution.pythonModelInstance.singleObservationContext.type != "time":
            raise ValueError(
                "The execution has to be in single observation mode on time data"
            )

    # Start date
    start_date = (
        model_execution.lastSuccessfulExecutionInstant
        - model_execution.pythonModelInstance.singleObservationContext.overlappingPeriod
    )

    # End date
    end_date = model_execution.currentExecutionInstant

    # Aggregation
    if (
        model_execution.pythonModelInstance.singleObservationContext.aggregationPeriod
        == "PT0S"
    ):
        aggregation = "RAW_VALUES"
    else:
        aggregation = "TIME"

    # Aggregation period
    if aggregation == "RAW_VALUES":
        aggregation_period = None
    else:
        aggregation_period = (
            model_execution.pythonModelInstance.singleObservationContext.aggregationPeriod
        )

    # Aggregation function
    if aggregation == "RAW_VALUES":
        aggregation_function = None
    else:
        aggregation_function = []
        for input_data in model_execution.get_data_input_dict(
            data_type="any", mode="object"
        ).values():
            if input_data is None:
                continue
            elif input_data.type in ["temporal-data-list", "batch-data-list"]:
                for data_value in input_data.dataList:
                    aggregation_function.append(data_value.aggregationFunction)
            else:
                aggregation_function.append(input_data.aggregationFunction)

    # Get data references' dict(sourceCode: ref) for all data
    input_data_references = model_execution.get_data_input_dict(
        data_type="any", mode="reference"
    )

    # Get data references' dict(sourceCode: ref) of scalar data
    input_scalar_data_objects = model_execution.get_data_input_dict(
        data_type="any", values_type="scalar", mode="object"
    )

    # list of references and unit-ids for scalar data
    scalar_data_references = []
    scalar_input_unit_ids = []
    for value in input_scalar_data_objects.values():
        if value is None:
            continue
        elif value.type in ["temporal-data-list", "batch-data-list"]:
            # add references to list
            scalar_data_references.extend(
                [input_data.reference for input_data in value.dataList]
            )
            # add unit-ids to list
            scalar_input_unit_ids.extend(
                [input_data.unit.id for input_data in value.dataList]
            )
        else:
            # add reference to list
            scalar_data_references.append(value.reference)
            # add unit-id to list
            scalar_input_unit_ids.append(value.unit.id)

    # Query endpoint
    data = api.get_multiple_data_values(
        start_date=start_date,
        end_date=end_date,
        aggregation=aggregation,
        data_id=None,
        data_reference=scalar_data_references,
        number_of_values=None,
        aggregation_period=aggregation_period,
        aggregation_function=aggregation_function,
        unit_id=scalar_input_unit_ids,
        unit_label=None,
        name_data_from="reference",
        append_unit_to_description=False,
        join_series_on="timestamp",
        api_credentials=api_credentials,
    )

    if rename_data_to_source_code_name:
        for k, v in input_data_references.items():
            value = input_scalar_data_objects.get(k)
            if value is not None and value.type in [
                "temporal-data-list",
                "batch-data-list",
            ]:
                values = []
                for index in data.index:
                    ser = pd.Series(
                        data=[data.loc[index, col] for col in v],
                        index=v,
                    )
                    ser.index.name = "reference"
                    values.append(ser)
                data[k] = pd.Series(data=values, index=data.index)
                data = data.drop(columns=v)
    # initialise list of DataFrames and Series to be concatenated
    dfs_list = [data]

    for input_data in model_execution.get_data_input_dict(
        data_type="time", values_type="vector", mode="object"
    ).values():
        vector_data = api.get_vector_time_values(
            vector_data_id=input_data.id,
            start_date=start_date,
            end_date=end_date,
            aggregation=aggregation,
            index_aggregation="RAW_VALUES",
            aggregation_period=aggregation_period,
            max_number_of_points=None,
            aggregation_function=aggregation_function,
            index_aggregation_function=None,
            index_aggregation_step_type=None,
            index_aggregation_step_value=None,
            min_index=None,
            max_index=None,
            index_unit_id=input_data.indexUnit.id,
            value_unit_id=input_data.valueUnit.id,
        )
        # vector data is converted to a list of series
        ser_list = [
            pd.Series(data=vector_data.loc[index]) for index in vector_data.index
        ]
        # list of series is converted to a series
        ser_vector_data = pd.Series(
            data=ser_list,
            index=vector_data.index,
            name=input_data.reference,
        )
        # Series of Series in added to the list of DataFrames for a later concatenation
        dfs_list.append(ser_vector_data)

    # list containing DataFrame and Series is concatenated
    data = pd.concat(dfs_list, axis=1)

    # Rename data
    if rename_data_to_source_code_name:
        # renaming dict(ref: sourceCode) for all data
        input_data_references = {
            k: v
            for k, v in input_data_references.items()
            if (
                input_scalar_data_objects.get(k) is None
                or input_scalar_data_objects.get(k).type
                not in ["temporal-data-list", "batch-data-list"]
            )
        }
        input_renaming_dict = utils.reverse_dict(input_data_references)
        data = data.rename(columns=input_renaming_dict)
    # Output
    return data


def get_single_obs_batch_data(
    model_execution: Optional[_dtos.ModelExecution] = None,
    rename_data_to_source_code_name: bool = True,
    api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    # Get variables from environment if not specified
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()
        if model_execution.pythonModelInstance.dataExchangeMode != "SINGLE_OBSERVATION":
            raise ValueError("The execution has to be in single observation mode")
        if model_execution.pythonModelInstance.singleObservationContext.type != "batch":
            raise ValueError(
                "The execution has to be in single observation mode on batches"
            )

    # Start date
    start_date = (
        model_execution.lastSuccessfulExecutionInstant
        - model_execution.pythonModelInstance.singleObservationContext.overlappingPeriod
    )

    # End date
    end_date = model_execution.currentExecutionInstant

    # Batch type
    batch_type_id = (
        model_execution.pythonModelInstance.singleObservationContext.batchPredicate.batchType.id
    )

    # Batch feature filters
    features_value_ids = [
        feature.id
        for feature in model_execution.pythonModelInstance.singleObservationContext.batchPredicate.featureFilters
    ]

    # Get data references' dict(sourceCode: ref) of scalar data
    input_scalar_data_objects = model_execution.get_data_input_dict(
        data_type="batch", values_type="scalar", mode="object"
    )

    # list of references and units of scalar data
    scalar_data_references = []
    scalar_input_unit_ids = []
    for value in input_scalar_data_objects.values():
        if value is None:
            continue
        elif value.type == "batch-data-list":
            # add references to the list
            scalar_data_references.extend(
                [input_data.reference for input_data in value.dataList]
            )
            # add unit-ids to the list
            scalar_input_unit_ids.extend(
                [input_data.unit.id for input_data in value.dataList]
            )
        else:
            # add reference to the list
            scalar_data_references.append(value.reference)
            # add unit to the list
            scalar_input_unit_ids.append(value.unit.id)

    # Get data references' dict(sourceCode: ref) for batch data (scalar and vector)
    input_data_references = model_execution.get_data_input_dict(
        data_type="batch", mode="reference"
    )

    # Get data references' dict(sourceCode: ref) for batch scalar data
    scalar_data_references = []
    for value in input_scalar_data_objects.values():
        if value is None:
            continue
        elif value.type == "batch-data-list":
            scalar_data_references.extend(
                [input_data.reference for input_data in value.dataList]
            )
        else:
            scalar_data_references.append(value.reference)

    # Query endpoint for data
    data = None

    if len(model_execution.get_data_input_dict()) > 0:
        data = api.get_multiple_data_values(
            start_date=start_date,
            end_date=end_date,
            aggregation="RAW_VALUES",
            data_reference=scalar_data_references,
            unit_id=scalar_input_unit_ids,
            name_data_from="reference",
            append_unit_to_description=False,
            join_series_on="index",
            api_credentials=api_credentials,
        )

    if rename_data_to_source_code_name:
        for k, v in input_data_references.items():
            if isinstance(v, list):
                values = []
                for index in data.index:
                    ser = pd.Series(
                        data=[data.loc[index, col] for col in v],
                        index=v,
                    )
                    ser.index.name = "reference"
                    values.append(ser)
                data[k] = pd.Series(data=values, index=data.index)
                data = data.drop(columns=v)

    # Query endpoint for features & batch predicate
    features = api.get_batches(
        batch_type_id=batch_type_id,
        start_date=start_date,
        end_date=end_date,
        features_value_ids=features_value_ids,
        api_credentials=api_credentials,
    )[1]

    # Rename features
    features_renaming_dict = utils.reverse_dict(
        model_execution.get_input_dict(input_types=["BATCH_TAG_KEY"], mode="reference")
    )

    features = features[
        [col for col in features.columns if col in features_renaming_dict.keys()]
    ]

    if rename_data_to_source_code_name is True:
        features = features.rename(columns=features_renaming_dict)

    if data is None:
        data = features
    else:
        data = data.join(features, how="right")

    # initialise list of DataFrames and Series to be concatenated
    dfs_list = [data]

    for input_data in model_execution.get_data_input_dict(
        data_type="batch", values_type="vector", mode="object"
    ).values():
        vector_data = api.get_vector_batch_values(
            vector_data_id=input_data.id,
            start_date=start_date,
            end_date=end_date,
            index_unit_id=input_data.indexUnit.id,
            value_unit_id=input_data.valueUnit.id,
            aggregation="RAW_VALUES",
            index_aggregation="RAW_VALUES",
            aggregation_period="PT0S",
            max_number_of_points=None,
            aggregation_function=None,
            index_aggregation_function=None,
            index_aggregation_step_type=None,
            index_aggregation_step_value=None,
            min_index=None,
            max_index=None,
            batch_ids=None,
        )
        # vector data is converted to a list of series
        ser_list = [
            pd.Series(data=vector_data.loc[index]) for index in vector_data.index
        ]
        # list of series is converted to a series
        ser_vector_data = pd.Series(
            data=ser_list,
            index=vector_data.index,
            name=input_data.reference,
        )
        # series of series in added to the list of DataFrames for a later concatenation
        dfs_list.append(ser_vector_data)

    # list containing 'data' with scalar values and the series with vector values is concatenated
    data = pd.concat(dfs_list, axis=1)

    # Rename data
    if rename_data_to_source_code_name:
        input_data_references = {
            k: v
            for k, v in input_data_references.items()
            if (
                input_scalar_data_objects.get(k) is None
                or input_scalar_data_objects.get(k).type != "batch-data-list"
            )
        }
        # renaming dict(ref: sourceCode) for batch data (scalar and vector)
        input_renaming_dict = utils.reverse_dict(input_data_references)
        data = data.rename(columns=input_renaming_dict)

    # Output
    return data



from typing import Optional, List, Union
from datetime import datetime

import requests

from .. import _credentials
from .. import utils


__all__ = [
    "get_batch_types",
    "get_batch_type_details",
    "get_single_batch",
    "get_batches",
    "update_batch_value",
    "update_batch_feature_value",
    "update_batch_features_and_values",
    "update_vector_batch_values",
    "create_batch",
    "update_batch",
    "create_or_update_batches",
    "delete_batch",
    "get_batch_relations",
    "get_single_batch_relation",
]


# Batches
def get_batch_types(
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types"
    response = requests.get(
        url=url,
        params={
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_batch_type_details(
    batch_type_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}"
    response = requests.get(url=url, **api_credentials.auth_kwargs)

    # Output
    response.raise_for_status()
    return response.json()


def get_single_batch(
    batch_type_id: str,
    batch_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches/{batch_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_batches(
    batch_type_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    name: Optional[str] = None,
    features_value_ids: Optional[Union[str, List[str]]] = None,
    asset_ids: Optional[Union[str, List[str]]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Format dates
    start_date_iso = utils.get_zulu_isoformat(start_date)
    end_date_iso = utils.get_zulu_isoformat(end_date)

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches"
    response = requests.get(
        url=url,
        params={
            "start": start_date_iso,
            "end": end_date_iso,
            "name": name,
            "feature-values": utils.join_list_query_param(features_value_ids),
            "asset-id": utils.join_list_query_param(asset_ids),
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_batch_value(
    batch_type_id: str,
    batch_id: str,
    data_id: str,
    value: Union[int, float],
    unit_id: Optional[str] = None,
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches/{batch_id}/value"
    response = requests.put(
        url=url,
        json={
            "dataId": data_id,
            "value": value,
            "unitId": unit_id,
        },
        params={"create-upload-event": create_upload_event},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def update_batch_feature_value(
    batch_type_id: str,
    batch_id: str,
    feature_id: str,
    value: str,
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches/{batch_id}/feature"
    response = requests.put(
        url=url,
        json={
            "batchTagKeyId": feature_id,
            "batchTagValueValue": value,
        },
        params={"create-upload-event": create_upload_event},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def update_batch_features_and_values(
    batch_type_id: str,
    data: List[dict],
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches/features-and-values"
    response = requests.put(
        url=url,
        json=data,
        params={"create-upload-event": create_upload_event},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def update_vector_batch_values(
    data: List[dict],
    use_external_reference: bool = False,
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/value-upload/vector-batch-values"
    response = requests.post(
        url=url,
        json=data,
        params={
            "use-external-reference": use_external_reference,
            "create-upload-event": create_upload_event,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def create_batch(
    batch_type_id: str,
    name: str,
    steps: List[dict],
    tag_values_by_id: Optional[List[str]] = None,
    tag_values_by_values: Optional[List[dict]] = None,
    values: Optional[List[dict]] = None,
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Init
    if tag_values_by_id is None:
        tag_values_by_id = []

    if tag_values_by_values is None:
        tag_values_by_values = []

    if values is None:
        values = []

    # Build payload
    batch_data = {
        "name": name,
        "steps": steps,
        "tagValuesById": tag_values_by_id,
        "tagValuesByValue": tag_values_by_values,
        "values": values,
    }

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches"
    response = requests.post(
        url=url,
        json=batch_data,
        params={"create-upload-event": create_upload_event},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_batch(
    batch_type_id: str,
    batch_id: str,
    name: str,
    steps: List[dict],
    tag_values_by_id: Optional[List[str]] = None,
    tag_values_by_values: Optional[List[dict]] = None,
    values: Optional[List[dict]] = None,
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing batch.

    Parameters
    ----------
    batch_type_id : str
        The OIAnalytics ID of the batch type the batch belongs to.
    batch_id : str
        The OIAnalytics ID of the batch to update. Must be an existing ID of a batch.
    name : str
        The new name of the batch.
    steps : list of dict
        The array of steps the batch has been through.
        These steps are merged with existing step.
        Any existing step on a given step id will be overwritten by the command.
        Each step is a dict with the following keys:
        - stepId: str; the id of the step in the batch type.
        - start: str; the start of the step in ISO format.
        - end : str; the end of the step in ISO format.
        - localisationType : {'NO_LOCALISATION', 'TAG_VALUES', 'ASSET'}; the type of localisation for this step.
        - localisationAssetId : list of str; the ids of the asset used to localise the step.
        - localisationTagValueIds : list of str; the ids of the tag value used to localise the step.
    tag_values_by_id : list of str, optional
        The array of batch feature id to create on the batch.
        This is an array of string.
        It is merged with existing feature values on an already existing feature id will be overwritten.
    tag_values_by_values : list of dict
        The array of batch feature values to add to the batch.
        It is merged with existing feature values on an already existing feature id will be overwritten.
        Each dict has the following keys:
        - batchTagKeyId: str; the ID of the feature.
        - batchTagValueValue: str; the value of the feature.
    values : list of dict
        The array of values to create on the batch.
        It is merged with existing feature values on an already existing feature id will be overwritten.
        Each dict has the following keys:
        - dataId: str; the ID of the data.
        - value: float | int, optional; the value (if None, the data will be deleted if it previously existed).
        - unitId: str; the ID of the unit the value is expressed in (in its absent, the storage unit of the measurement will be used).
    create_upload_event : bool, default True
        Whether to create an upload event with updated values.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API.
        If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Init
    if tag_values_by_id is None:
        tag_values_by_id = []

    if tag_values_by_values is None:
        tag_values_by_values = []

    if values is None:
        values = []

    # Build payload
    batch_data = {
        "name": name,
        "steps": steps,
        "tagValuesById": tag_values_by_id,
        "tagValuesByValue": tag_values_by_values,
        "values": values,
    }

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches/{batch_id}"
    response = requests.put(
        url=url,
        json=batch_data,
        params={"create-upload-event": create_upload_event},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def create_or_update_batches(
    batch_type_id: str,
    batches: List[dict],
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches/create-or-update"
    response = requests.post(
        url=url,
        json=batches,
        params={"create-upload-event": create_upload_event},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def delete_batch(
    batch_type_id: str,
    batch_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> None:
    """
    Delete a batch identified by its id.

    Parameters
    ----------
    batch_type_id : str
        The OIAnalytics ID of the batch type the batch belongs to.
    batch_id : str
        The OIAnalytics ID of the batch to set the value to. Must be an existing id of a batch.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    None

    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches/{batch_id}"
    requests.delete(
        url=url,
        **api_credentials.auth_kwargs,
    )


# Batch Relations
def get_batch_relations(
    batch_structure_relation_id: Optional[str] = None,
    start_date: Optional[Union[str, datetime]] = None,
    end_date: Optional[Union[str, datetime]] = None,
    source_batch_name: Optional[str] = None,
    destination_batch_name: Optional[str] = None,
    page: int = 0,
    page_size: int = 20,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Format dates
    start_date_iso = utils.get_zulu_isoformat(start_date)
    end_date_iso = utils.get_zulu_isoformat(end_date)

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-relations"
    response = requests.get(
        url=url,
        params={
            "batchStructureRelationId": batch_structure_relation_id,
            "start": start_date_iso,
            "end": end_date_iso,
            "sourceBatchName": source_batch_name,
            "destinationBatchName": destination_batch_name,
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_single_batch_relation(
    batch_relation_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-relations/{batch_relation_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



from typing import Optional, Union, Tuple, Literal

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils


__all__ = ["get_model_instances", "get_single_model_instance"]


def get_model_instances(
    page: int = 0,
    page_size: int = 20,
    get_all_pages: bool = True,
    expand_creator: bool = True,
    expand_trigger: bool = True,
    expand_python_model: bool = True,
    multithread_pages: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """

    Parameters
    ----------
    page : int, default 0
        The requested page. It is ignored if 'get_all_pages' is set to True.
    page_size : int, default 20
        The size of the pages. It is ignored if 'get_all_pages' is set to True.
    get_all_pages : bool, default True
        Whether to list all model instances of all pages.
        If the value is set to True, the 'page' and 'page_size' arguments will be ignored, and the function will retrieve all pages.
    multithread_pages : bool, default True
        Whether to retrieve pages in multiple threads simultaneously.
        It is used only if 'get_all_pages' is set to True.
    expand_creator : bool, default True
        Whether to split the 'creator' column.
        If True, it is split into two columns: 'creator_id' and 'creator_friendlyName'.
    expand_trigger : bool, default True
        Whether to split the 'trigger' column.
        If True, it is split into two columns: 'trigger_type' and 'trigger_cron'.
    expand_python_model : bool, default True
        Whether to split the 'pythonModel' column.
        If True, it is split into two columns: 'pythonModel_id', 'pythonModel_name' and 'pythonModel_version'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.DataFrame

    """

    if get_all_pages:
        page = 0
        page_size = 500

    def get_page(page_num: int) -> dict:
        return endpoints.model_instances.get_model_instances(
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )

    def parse_page(page_response: dict) -> pd.DataFrame:
        page_df = pd.DataFrame(page_response["content"])
        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=[
                    "id",
                    "title",
                    "description",
                    "creator",
                    "creationDate",
                    "trigger," "active",
                    "lastSuccessfulExecution",
                    "pythonModel",
                    "upToDate",
                ],
            )
        # Set index
        page_df.set_index("id", inplace=True)
        return page_df

    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    if expand_creator:
        df = utils.expand_dataframe_column(
            df=df,
            col="creator",
        )

    if expand_trigger:
        df = utils.expand_dataframe_column(
            df=df,
            col="trigger",
        )

    if expand_python_model:
        df = utils.expand_dataframe_column(
            df=df,
            col="pythonModel",
        )
    # Output
    return df


def get_single_model_instance(
    model_instance_id: str,
    expand_creator: bool = True,
    expand_trigger: bool = True,
    expand_python_model: bool = True,
    create_parameters_dataframes: bool = False,
    extract_from_input_parameters: Optional[Literal["id", "reference"]] = "reference",
    extract_from_output_parameters: Optional[Literal["id", "reference"]] = "reference",
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> Union[pd.Series, Tuple[pd.Series, pd.DataFrame, pd.DataFrame]]:
    """

    Parameters
    ----------
    model_instance_id : str
        The ID of the retrieved python model instance.
    expand_creator : bool, default True
        Whether to split the 'creator' column.
        If True, it is split into two columns: 'creator_id' and 'creator_friendlyName'.
    expand_trigger : bool, default True
        Whether to split the 'trigger' column.
        If True, it is split into two columns: 'trigger_type' and 'trigger_cron'.
    expand_python_model : bool, default True
        Whether to split the 'pythonModel' column.
        If True, it is split into two columns: 'pythonModel_id', 'pythonModel_name' and 'pythonModel_version'.
    create_parameters_dataframes : bool, default False
        Create additional DataFrames with input-parameters and output-parameters.
        If True, the response is a tuple with:
          - one Series (model instance);
          - two DataFrames (the first one for the input parameters and the second for the output parameters).
    extract_from_input_parameters : {'id', 'reference', None}, default 'reference'
        Value to be exposed in the 'value' column. Used only if 'create_parameters_dataframes' is True.
    extract_from_output_parameters : {'id', 'reference', None}, default 'reference'
        Value to be exposed in the 'value' column. Used only if 'create_parameters_dataframes' is True.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series | tuple[pd.Series, pd.DataFrame, pd.DataFrame]
    """
    response = endpoints.model_instances.get_single_model_instance(
        model_instance_id=model_instance_id, api_credentials=api_credentials
    )
    df = pd.DataFrame([response])

    if len(df) == 0:
        if create_parameters_dataframes:
            return df.squeeze(), pd.DataFrame(), pd.DataFrame()
        else:
            return df.squeeze()

    df.set_index("id", inplace=True)

    if expand_creator:
        df = utils.expand_dataframe_column(
            df=df,
            col="creator",
        )

    if expand_trigger:
        df = utils.expand_dataframe_column(
            df=df,
            col="trigger",
        )

    if expand_python_model:
        df = utils.expand_dataframe_column(
            df=df,
            col="pythonModel",
        )

    if not create_parameters_dataframes:
        return df.squeeze()

    def expand_dict(d: dict) -> pd.DataFrame:
        if isinstance(d["value"], dict):
            d = d | {f"value_{key}": value for key, value in d["value"].items()}
            d.pop("value")
        df_expanded = pd.DataFrame([d])
        df_expanded.set_index("sourceCodeName", inplace=True)
        return df_expanded

    if extract_from_input_parameters is not None:
        for parameter in response["inputParameters"]:
            if isinstance(parameter["value"], dict):
                parameter["value"] = parameter["value"][extract_from_input_parameters]
        df_input = pd.DataFrame(response["inputParameters"])
        df_input.set_index("sourceCodeName", inplace=True)
    else:
        df_input = pd.concat(
            [expand_dict(parameter) for parameter in response["inputParameters"]],
            axis=0,
        )

    if extract_from_output_parameters is not None:
        for parameter in response["outputParameters"]:
            if isinstance(parameter["value"], dict):
                parameter["value"] = parameter["value"][extract_from_output_parameters]
        df_output = pd.DataFrame(response["outputParameters"])
        df_output.set_index("sourceCodeName", inplace=True)
    else:
        df_output = pd.concat(
            [expand_dict(parameter) for parameter in response["outputParameters"]],
            axis=0,
        )

    return df.squeeze(), df_input, df_output
